﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using MVR.FileManagementSecure;
using System.Text.RegularExpressions;
using System.Linq;
using Leap.Unity.Query;
using Leap.Unity;
using MVR.Hub;
using UnityEngine.UI;
using Oculus.Platform;

namespace DLPlugin
{
    /*
    DependencyViewer is a script for finding all dependencies in the scene, and their sub dependencies.
    You can scan current scene or select VAM preset (.vap)/.json file. Select dependency from list of root dependencies, and look each case of its usage.
    You can download Hub dependencies for ANY package (even Paid)!

    By DemonLord. License [CC BY] v2.3
    */
    public class DependencyViewer : MVRScript {

        private JSONStorableStringChooser deps;

        private RectTransform fullWidthUIContent;
        private List<Transform> fullWidthUIElements;
        private UIDynamicButton loadPresetButton;
        private UIDynamicPopup depsUI;
        private UIDynamicTextField tfUI;

        private string[] packages;
        private string selfName;
        private string allDep;
        private string fileName;
        private string filePackageName;
        private JSONClass jsNow;
        private Dictionary<string, List<string>> paths = new Dictionary<string, List<string>>();
        private Dictionary<string, VarItem> varNames = new Dictionary<string, VarItem>();
        private bool fromMetaJson;
        private float wait;
        private HubDownloader packageDownloader;
        private GameObject myPackageDownloaderUIObject;
        private HubDownloaderUI hubDownloader;
        private HubDownloaderUI myHubDownloader;
        private float counter = 0;
        private List<Slider> sliders = new List<Slider>();
        private bool scanned;
        private int toFind;
        private Text myHubDownloaderHeaderText;
        private int pendingResourceDownloads = 0;
        private string lastPath = "";

        public struct VarItem
        {
            public string name;
            public string license;
            public bool isLatest;
            public string scenePath;
            public bool havePackage;
            public bool isSubDependency;
            public HubResourcePackageUI uiNew;
            public HubResourcePackageUI uiFinded;
        }
        public override void Init() {
			try
            {
                //SuperController.LogMessage("DependencyViewer ");
                packageDownloader = SuperController.singleton.packageDownloader;
                hubDownloader = SuperController.singleton.transform.Find("WorldScaleAdjust/HUD/Browsers/Canvas/HubDownloaderUI").GetComponent<HubDownloaderUI>();

                fullWidthUIElements = new List<Transform>();

                loadPresetButton = CreateButton("Load Json", true);
                loadPresetButton.transform.SetAsFirstSibling();
                loadPresetButton.button.onClick.AddListener(() =>
                {
                    List<ShortCut> shortCuts = new List<ShortCut>();
                    ShortCut shortCut = new ShortCut();
                    shortCut.displayName = "Saves\\scene";
                    shortCut.path = "Saves\\scene";
                    shortCuts.Add(shortCut);
                    shortCut = new ShortCut();
                    shortCut.displayName = "Custom";
                    shortCut.path = "Custom";
                    shortCuts.Add(shortCut);
                    shortCut = new ShortCut();
                    shortCut.displayName = "Custom\\Atom\\Person";
                    shortCut.path = "Custom\\Atom\\Person";
                    shortCuts.Add(shortCut);
                    shortCuts.AddRange(FileManagerSecure.GetShortCutsForDirectory("", false, false, false, false));
                    SuperController.singleton.GetMediaPathDialog((string s) => LoadJson(s), "json|vap", "Saves", false, true, false, null, false, shortCuts, true, true);
                });
                ColorBlock colors = loadPresetButton.button.colors;
                colors.normalColor = new Color(0.5f, 1f, 0.5f, 1);
                colors.highlightedColor = new Color(0.4f, 0.8f, 0.4f, 1);
                colors.pressedColor = new Color(0.3f, 0.5f, 0.3f, 1);
                loadPresetButton.button.colors = colors;

                UIDynamicButton b1 = CreateButton("Check Current scene");
				b1.button.onClick.AddListener(CheckCurrentScene);
                b1.button.colors = colors;

                UIDynamicButton b2 = CreateButton("Rescan");
                b2.button.onClick.AddListener(RescanCurrent);
                colors.normalColor = new Color(1f, 1f, 0.5f, 1);
                colors.highlightedColor = new Color(0.8f, 0.8f, 0.5f, 1);
                colors.pressedColor = new Color(0.7f, 0.7f, 0.4f, 1);
                b2.button.colors = colors;

                UIDynamicButton b3 = CreateButton("Check Download Dependencies",true);
                b3.button.onClick.AddListener(CheckDownloadDependencies);
                colors.normalColor = new Color(1f, 0.6f, 0.9f, 1);
                colors.highlightedColor = new Color(0.8f, 0.4f, 0.7f, 1);
                colors.pressedColor = new Color(0.6f, 0.3f, 0.55f, 1);
                b3.button.colors = colors;
            }
			catch (Exception e) {
				SuperController.LogError("DependencyViewer: " + e);
			}
        }
        public override void InitUI()
        {
            base.InitUI();
            if (!(UITransform != null))
            {
                return;
            }

            MVRScriptUI componentInChildren = UITransform.GetComponentInChildren<MVRScriptUI>();
            if (componentInChildren.fullWidthUIContent == null)
            {
                return;
            }

            fullWidthUIContent = componentInChildren.fullWidthUIContent;
            foreach (Transform fullWidthUIElement in fullWidthUIContent)
            {
                fullWidthUIElement.gameObject.SetActive(value: true);
                fullWidthUIElement.SetParent(componentInChildren.fullWidthUIContent, worldPositionStays: false);
            }


            List<string> list2 = new List<string> { "Start Info" };
            deps = new JSONStorableStringChooser("Dependencies", list2, "Start Info", "Dependencies", SelectDependency);
            depsUI = CreateFullWidthFilterablePopup(deps);
            depsUI.labelWidth *= 1.3f;

            allDep = "<b>DependencyViewer</b> is a script for finding all dependencies in the scene, and their sub dependencies. And downloading dependencies from selected scene, or from meta.json file in package.\n\n" +
                "You can <color=green>scan current scene</color> or <color=green>load</color> VAM preset<color=brown>.vap</color> , scene<color=brown>.json</color> or <color=brown>meta.json</color> file from package.\n" +
                "Select dependency from list of root dependencies, and look each case of its usage. (Plugin traces each occurrence of \":/\" string in json code.)\n\n" +
                "<b>To remove sub dependency</b> you need to remove their root dependency, because only it appears in scene.\n\n" +
                "<b>All Dependencies</b> tab has root dependencies, and generate their sub dependencies iteratively" +
                " (it looks inside meta.json file of that package, and for each of its <b>main</b> dependencies, looks inside their package recursively).\n\n" +
                "<b>You can download</b> Hub dependencies for ANY package (<b>even Paid</b>)!\n" +
                "(<color=green>Load json</color> -> select package -> <color=brown>meta.json</color>)\n\n" +
                "When you scan <color=brown>meta.json</color> file, dependencies will be shown exactly from that file. " +
                "And when downloading [<color=purple>NOTFOUND</color>] packages, their sub dependencies <b>will not</b> be shown!\n\n" +
                "For [<color=purple>V.1->2</color>] case, will try download <b>old</b> version, because scanned file require <b>exact</b> version, <b>not latest</b> (mainly because of scripts)." +
                "\n\nBy <color=purple>DemonLord</color> [<color=green>CC-BY</color>] v2.3";
            tfUI = CreateFullWidthTextField(new JSONStorableString("All Dependencies", allDep));
            tfUI.height = 990;
            //SuperController.LogMessage(allDep);

        }
        void Update()
        {
            if (myHubDownloader != null && myHubDownloader.panel.gameObject.activeSelf)
            {
                if (myHubDownloader.downloadInfoText != null)
                {
                    if (pendingResourceDownloads > 0)
                    {
                        myHubDownloader.downloadInfoText.text = "Queued downloads count: " + pendingResourceDownloads;
                    }
                    else
                    {
                        myHubDownloader.downloadInfoText.text = "No queued downloads";
                    }
                }
            }
            if (wait != 0 )
            {
                if (wait + 1 < Time.time)
                {
                    wait = Time.time;
                    counter++;
                    if (myHubDownloaderHeaderText != null)
                    {
                        myHubDownloaderHeaderText.text = "Packages (" + myHubDownloader.GetComponentsInChildren<HubResourcePackageUI>().Length + ")";
                    }
                    HubResourcePackageUI[] c2 = hubDownloader.GetComponentsInChildren<HubResourcePackageUI>(true);
                    if (c2 != null)
                    {
                        try
                        {
                            string nameV;
                            string[] spl;
                            //SuperController.LogMessage("c2.Length " + c2.Length);
                            JSONClass js = new JSONClass();
                            js["filename"] = "a.b.1.var";
                            VarItem vi;
                            bool needSubDep = false;
                            int lastNeedSubDepSiblingIndex = 0;
                            foreach (HubResourcePackageUI ui in c2)
                            {
                                if (!ui.gameObject.activeSelf) continue;
                                if (needSubDep)
                                {
                                    if (ui.isDependencyIndicator.activeSelf)
                                    {
                                        ui.transform.SetParent(myHubDownloader.packagesContainer, false);
                                        ui.transform.SetSiblingIndex(lastNeedSubDepSiblingIndex);
                                        ui.licenseTypeText.text = Colorize(ui.licenseTypeText.text).Replace("green", "lime").Replace("red", "#E70000");
                                        ui.fileSizeText.text = "<color=white>" + ui.fileSizeText.text + "</color>";
                                        lastNeedSubDepSiblingIndex++;
                                        continue;
                                    }
                                    else 
                                        needSubDep = false;
                                }
                                nameV = ui.nameText.text;
                                spl = nameV.Split('.');
                                if (spl.Length != 3) continue;
                                nameV = spl[0] + '.' + spl[1] + '.' + "latest";
                                if (varNames.ContainsKey(nameV) && !ui.nameText.text.EndsWith("latest"))
                                {
                                    vi = varNames[nameV];
                                    varNames.Remove(nameV);
                                    varNames[ui.nameText.text] = vi;
                                }
                                nameV = ui.nameText.text;

                                if (varNames.ContainsKey(nameV))
                                {
                                    vi = varNames[nameV];
                                    if (vi.uiFinded != null)
                                    {
                                        if (vi.uiFinded.nameText.text.EndsWith("latest") && vi.uiFinded.nameText.text != nameV)
                                        {
                                            int n = vi.uiFinded.transform.GetSiblingIndex();
                                            ui.transform.SetParent(myHubDownloader.packagesContainer, false);
                                            vi.uiFinded.connectedItem = new HubResourcePackage(js, null, false);
                                            vi.uiFinded.gameObject.SetActive(value: false);
                                            ui.transform.SetSiblingIndex(n);
                                            ui.resourceButton.interactable = true;
                                            Button button = Instantiate(ui.resourceButton, ui.resourceButton.transform.parent, false);
                                            Destroy(ui.resourceButton.gameObject);
                                            ui.resourceButton = button;
                                            ui.resourceButton.onClick.AddListener(() => {
                                                SuperController.singleton.hubBrowser.OpenDetail(ui.nameText.text, true);
                                                //SuperController.LogMessage("nameV2 " + ui.nameText.text);
                                            });

                                            ui.connectedItem.downloadQueuedCallback = (HubResourcePackage.DownloadQueuedCallback)Delegate.Combine(new HubResourcePackage.DownloadQueuedCallback(ResourceDownloadQueued), ui.connectedItem.downloadQueuedCallback);
                                            ui.connectedItem.downloadCompleteCallback = (HubResourcePackage.DownloadCompleteCallback)Delegate.Combine(new HubResourcePackage.DownloadCompleteCallback(ResourceDownloadComplete), ui.connectedItem.downloadCompleteCallback);
                                            ui.connectedItem.downloadErrorCallback = (HubResourcePackage.DownloadErrorCallback)Delegate.Combine(new HubResourcePackage.DownloadErrorCallback(ResourceDownloadError), ui.connectedItem.downloadErrorCallback);

                                            ui.licenseTypeText.text = Colorize(ui.licenseTypeText.text).Replace("green", "lime").Replace("red", "#E70000");
                                            ui.fileSizeText.text = "<color=white>" + ui.fileSizeText.text + "</color>";
                                            ui.isDependencyIndicator.SetActive(varNames[nameV].isSubDependency);
                                            vi.uiFinded = ui;
                                            varNames[nameV] = vi;
                                        }
                                        else if (vi.uiFinded.nameText.text == nameV && ui.transform.parent == hubDownloader.packagesContainer)
                                        {
                                            ui.connectedItem = new HubResourcePackage(js, null, false);
                                            ui.gameObject.SetActive(value: false);
                                        }
                                    }
                                    
                                    if (vi.uiNew != null)
                                    {
                                        toFind--;
                                        int n = vi.uiNew.transform.GetSiblingIndex();
                                        ui.transform.SetParent(myHubDownloader.packagesContainer, false);
                                        Destroy(vi.uiNew.gameObject);
                                        vi.uiNew = null;
                                        ui.transform.SetSiblingIndex(n);
                                        ui.resourceButton.interactable = true;
                                        Button button = Instantiate(ui.resourceButton, ui.resourceButton.transform.parent, false);
                                        Destroy(ui.resourceButton.gameObject);
                                        ui.resourceButton = button;
                                        ui.resourceButton.onClick.AddListener(() => {
                                            SuperController.singleton.hubBrowser.OpenDetail(ui.nameText.text, true);
                                            //SuperController.LogMessage("nameV1 " + ui.nameText.text);
                                        });

                                        ui.connectedItem.downloadQueuedCallback = (HubResourcePackage.DownloadQueuedCallback)Delegate.Combine(new HubResourcePackage.DownloadQueuedCallback(ResourceDownloadQueued), ui.connectedItem.downloadQueuedCallback);
                                        ui.connectedItem.downloadCompleteCallback = (HubResourcePackage.DownloadCompleteCallback)Delegate.Combine(new HubResourcePackage.DownloadCompleteCallback(ResourceDownloadComplete), ui.connectedItem.downloadCompleteCallback);
                                        ui.connectedItem.downloadErrorCallback = (HubResourcePackage.DownloadErrorCallback)Delegate.Combine(new HubResourcePackage.DownloadErrorCallback(ResourceDownloadError), ui.connectedItem.downloadErrorCallback);

                                        ui.licenseTypeText.text = Colorize(ui.licenseTypeText.text).Replace("green", "lime").Replace("red", "#E70000");
                                        if (!ui.isDependencyIndicator.activeSelf)
                                        {
                                            if(!vi.havePackage && !fromMetaJson) 
                                            {
                                                needSubDep = true;
                                                lastNeedSubDepSiblingIndex = n+1;
                                                //SuperController.LogMessage("needSubDep " + ui.nameText.text);
                                            }
                                        }
                                        else
                                            ui.isDependencyIndicator.SetActive(value: false);
                                        ui.fileSizeText.text = "<color=white>" + ui.fileSizeText.text + "</color>";
                                        ui.isDependencyIndicator.SetActive(varNames[nameV].isSubDependency);
                                        vi.uiFinded = ui;
                                        varNames[nameV] = vi;
                                    }
                                }
                                else
                                {
                                    ui.connectedItem = new HubResourcePackage(js, null, false);
                                    ui.gameObject.SetActive(value: false);
                                }
                            }
                        }
                        catch (Exception e)
                        {
                            SuperController.LogMessage("Update " + e);
                        }
                    }
                    foreach (Slider s in sliders)
                    {
                        if (s != null)
                        {
                            s.value = counter / 20f;
                        }
                    }
                    if (toFind == 0)
                    {
                        toFind = -1;
                        counter = 17;
                    }
                }
                if (counter == 10)
                {
                    foreach (Slider s in sliders)
                    {
                        if (s != null)
                        {
                            s.transform.Find("Fill Area/Fill").GetComponent<Image>().color = new Color(1f, 0.478f, 0.478f, 1f);
                            s.transform.Find("Text").GetComponent<Text>().text = "<- Try open link";
                        }
                    }
                }
                if (counter >= 20)
                {
                    wait = 0;
                }
            }
        }
        protected void ResourceDownloadQueued(HubResourcePackage hrp)
        {
            pendingResourceDownloads++;
        }

        protected void ResourceDownloadComplete(HubResourcePackage hrp, bool alreadyHad)
        {
            pendingResourceDownloads--;
        }

        protected void ResourceDownloadError(HubResourcePackage hrp, string err)
        {
            pendingResourceDownloads--;
        }

        private string GetScenePath(string path)
        {
            foreach (string s2 in SuperController.singleton.GetFilesAtPath(path))
            {
                if (s2.EndsWith(".json")) return s2;
            }
            foreach (string s in FileManagerSecure.GetDirectories(path))
            {
                return GetScenePath(s);
            }
            return null;
        }
        private void CheckDownloadDependencies()
        {
            try
            {
                if (scanned)
                {
                    myPackageDownloaderUIObject.SetActive(true);
                    return;
                }
                if (varNames.Count > 0 && packageDownloader != null)
                {
                    scanned = true;
                    wait = Time.time;
                    sliders.Clear();
                    counter = 0;
                    packageDownloader.ClearTrackedPackages();
                    foreach (var item in varNames)
                    {
                        if (item.Value.uiNew != null)
                            Destroy(item.Value.uiNew.gameObject);
                    }
                    if (myHubDownloader != null)
                    {
                        foreach (var item in myHubDownloader.packagesContainer.GetComponents<HubResourcePackageUI>())
                        {
                            Destroy(item);
                        }
                    }
                    if (myPackageDownloaderUIObject == null)
                    {
                        HubDownloaderUI c0 = SuperController.singleton.GetComponentInChildren<HubDownloaderUI>(true);
                        myPackageDownloaderUIObject = Instantiate(c0.gameObject, this.manager.scriptUIParent, false);
                        myHubDownloaderHeaderText = myPackageDownloaderUIObject.transform.Find("InnerPanel/HubDownloads/DownloadsHeader/Package").GetComponent<Text>();
                        HubDownloaderUI c1 = myPackageDownloaderUIObject.GetComponentInChildren<HubDownloaderUI>();
                        if (c1 != null)
                        {
                            c1.panel = myPackageDownloaderUIObject.GetComponent<RectTransform>();
                            c1.panel.anchoredPosition = new Vector2(c1.panel.anchoredPosition.x, c1.panel.anchoredPosition.y + 50);
                            c1.panel.sizeDelta = new Vector2(c1.panel.sizeDelta.x, c1.panel.sizeDelta.y + 220);
                            c1.packagesContainer = c1.transform.Find("InnerPanel/HubDownloads/Downloads/Viewport/Content").GetComponent<RectTransform>();
                            c1.infoText = c1.transform.Find("InnerPanel/InfoText").GetComponent<Text>();
                            c1.downloadInfoText = c1.transform.Find("InnerPanel/DownloadInfoText").GetComponent<Text>();
                            /*
                            SuperController.LogMessage("c11 ");
                            c1.openPanelButton = CDD.button;
                            c1.openPanelButton.onClick.AddListener(() =>
                            {
                                if (c1.panel != null)
                                {
                                    c1.panel.gameObject.SetActive(value: true);
                                }
                            });*/
                            c1.closePanelButton = c1.transform.Find("InnerPanel/CloseButton").GetComponent<Button>();
                            c1.closePanelButton.onClick.AddListener(() =>
                            {
                                if (c1.panel != null)
                                {
                                    c1.panel.gameObject.SetActive(value: false);
                                }
                            });
                            c1.clearTrackedPackagesButton = c1.transform.Find("InnerPanel/ClearAllButton").GetComponent<Button>();
                            c1.clearTrackedPackagesButton.onClick.AddListener(() =>
                            {
                                packageDownloader.ClearTrackedPackages();
                                foreach (var item in varNames)
                                {
                                    if (item.Value.uiNew != null)
                                        Destroy(item.Value.uiNew.gameObject);
                                }
                                if (myHubDownloader != null)
                                {
                                    foreach (var item in myHubDownloader.packagesContainer.GetComponents<HubResourcePackageUI>())
                                    {
                                        Destroy(item);
                                    }
                                    myHubDownloaderHeaderText.text = "Packages (0)";
                                }
                                counter = 20;
                                scanned = false;
                            });
                            c1.packageNameInputField = c1.transform.Find("InnerPanel/InputField").GetComponent<InputField>();
                            if (c1.packageNameInputField != null)
                            {
                                c1.packageNameInputField.gameObject.SetActive(value: false);
                            }
                            c1.findPackageButton = c1.transform.Find("InnerPanel/FindPackageButton").GetComponent<Button>();
                            c1.findPackageButton.transform.Find("Text").GetComponent<Text>().text = "Rescan";
                            c1.findPackageButton.onClick.AddListener(() =>
                            {
                                scanned = false;
                                CheckDownloadDependencies();
                            });/*
                            if (c1.findPackageButton != null)
                            {
                                c1.findPackageButton.gameObject.SetActive(value: false);
                            }*/
                            c1.downloadAllTrackedPackagesButton = c1.transform.Find("InnerPanel/DownloadAllButton").GetComponent<Button>();
                            c1.downloadAllTrackedPackagesButton.onClick.AddListener(() =>
                            {
                                foreach (var item in varNames)
                                {
                                    if (item.Value.uiFinded != null)
                                    {
                                        SuperController.LogMessage("name " + item.Value.uiFinded.nameText + (item.Value.uiFinded.connectedItem != null));
                                        if (item.Value.uiFinded.connectedItem != null)
                                            item.Value.uiFinded.connectedItem.Download();
                                    }
                                        
                                }
                            });
                            c1.disabledIndicator = c1.transform.Find("PackageDownloaderDisabledIndicator").gameObject;
                            c1.enableHubDownloaderButton = c1.transform.Find("PackageDownloaderDisabledIndicator/AgreeButton").GetComponent<Button>();
                            c1.enableHubDownloaderButton.onClick.AddListener(() =>
                            {
                                if (UserPreferences.singleton != null)
                                {
                                    UserPreferences.singleton.enableHubDownloader = true;
                                    c1.disabledIndicator.SetActive(value: false);
                                }
                            });
                            c1.rejectHubDownloaderButton = c1.transform.Find("PackageDownloaderDisabledIndicator/CloseButton").GetComponent<Button>();
                            c1.rejectHubDownloaderButton.onClick.AddListener(() =>
                            {
                                if (c1.panel != null)
                                {
                                    c1.panel.gameObject.SetActive(value: false);
                                }
                            });

                            c1.transform.GetComponent<Image>().enabled = false;
                            c1.transform.Find("InnerPanel").GetComponent<Image>().color = new Color(0.5137869f, 0.06893385f, 0.625f, 1f);
                            c1.disabledIndicator.GetComponent<Image>().color = new Color(0.5137869f, 0.06893385f, 0.625f, 1f);
                            if (UserPreferences.singleton != null && UserPreferences.singleton.enableHubDownloader)
                            {
                                c1.disabledIndicator.SetActive(value: false);
                            }
                            myHubDownloader = c1;
                        }
                    }

                    myPackageDownloaderUIObject.SetActive(true);
                    VarItem vi;
                    toFind = varNames.Keys.Count;
                    string[] spl;
                    int version;
                    string name;
                    JSONClass js;
                    foreach (string key in varNames.Keys.ToArray())
                    {
                        packageDownloader.FindPackage(key);
                        RectTransform rectTransform = SuperController.singleton.hubBrowser.CreateDownloadPrefabInstance();
                        if (rectTransform != null)
                        {
                            rectTransform.SetParent(myHubDownloader.packagesContainer, worldPositionStays: false);
                            HubResourcePackageUI rpUI = rectTransform.GetComponent<HubResourcePackageUI>();
                            if (rpUI != null)
                            {
                                vi.scenePath = null;
                                vi = varNames[key];
                                vi.uiNew = rpUI;
                                rpUI.nameText.text = key;
                                rpUI.isDependencyIndicator.SetActive(varNames[key].isSubDependency);
                                rpUI.licenseTypeText.text = varNames[key].license.Replace("green", "lime").Replace("red", "#E70000");
                                rpUI.fileSizeText.text = "<color=white>null</color>";
                                rpUI.resourceButton.onClick.AddListener(() => { SuperController.singleton.hubBrowser.OpenDetail(key, true); });
                                rpUI.downloadButton.gameObject.SetActive(value: false);
                                spl = key.Split('.');
                                if (spl.Length != 3) continue;
                                version = FileManagerSecure.GetPackageVersion(key);
                                //SuperController.LogMessage("rpUI " + key + " " + version);
                                if (version > 0)
                                {
                                    name = spl[0] + "." + spl[1] + "." + version;
                                    try
                                    {
                                        if (FileManagerSecure.DirectoryExists("AddonPackages\\" + name + ".var:\\Saves\\scene")) 
                                        {
                                            vi.scenePath = GetScenePath("AddonPackages\\" + name + ".var:\\Saves\\scene");
                                        }
                                        
                                    }
                                    catch (Exception e)
                                    {
                                        SuperController.LogMessage("CheckDownloadDependencies : LoadJSON " + e);
                                    }
                                    if(vi.scenePath != null)
                                    {
                                        rpUI.openSceneButton.gameObject.SetActive(value: true);
                                        rpUI.openSceneButton.onClick.AddListener(() =>
                                        {
                                            if (myHubDownloader.panel != null)
                                            {
                                                myHubDownloader.panel.gameObject.SetActive(value: false);
                                            }
                                            SuperController.singleton.Load(varNames[key].scenePath);
                                        });
                                    }
                                    else
                                    {
                                        rpUI.openInPackageManagerButton.gameObject.SetActive(value: true);
                                        rpUI.openInPackageManagerButton.onClick.AddListener(() =>
                                        {
                                            if (FileManagerSecure.PackageExists(key))
                                            {
                                                if (myHubDownloader.panel != null)
                                                {
                                                    myHubDownloader.panel.gameObject.SetActive(value: false);
                                                }
                                                SuperController.singleton.OpenPackageInManager(key);
                                            }
                                        });
                                    }
                                }
                                else
                                {
                                    rpUI.progressSlider.gameObject.SetActive(value: true);
                                    sliders.Add(rpUI.progressSlider.GetComponent<Slider>());
                                    rpUI.progressSlider.transform.Find("Fill Area/Fill").GetComponent<Image>().color = new Color(0.84f, 0.84f, 0.84f, 1f);
                                    rpUI.progressSlider.transform.Find("Text").GetComponent<Text>().text = "Wait Hub URL";
                                }
                                varNames[key] = vi;
                            }
                        }
                    }
                    myHubDownloaderHeaderText.text = "Packages (" + varNames.Count + ")";
                    if (filePackageName != "")
                    {
                        packageDownloader.FindPackage(filePackageName);
                        packageDownloader.FindPackage(GetPacGroupUid(filePackageName));
                    }
                }
                else
                {
                    tfUI.text = allDep + "\n<color=purple>No dependencies to check.</color>";
                }
            }
            catch (Exception e)
            {
                SuperController.LogMessage("CheckDownloadDependencies " + e);
            }
        }
        private string GetPacGroupUid(string packageName)
        {
            packageName = packageName.Replace("AddonPackages\\", String.Empty);
            string[] spl = packageName.Split('.');
            if (spl.Length < 3)
            {
                //SuperController.LogError("GetPacGroupUid: {spl.Length < 3} packageName: " + packageName);
                return packageName;
            }
            return spl[0] + '.' + spl[1];
        }
        private void LoadJson(string path)
        {
            if (string.IsNullOrEmpty(path))
                return;
            lastPath = path;
            filePackageName = "";
            if (path.Contains(":/"))
            {
                string[] spl = path.Split(new string[] { ":/" }, StringSplitOptions.None);
                if (spl.Length > 1 && spl[0].Split('.').Length == 3)
                    filePackageName = spl[0];
                //SuperController.LogMessage("filePackageName " + filePackageName);
            }
            fileName = "[<color=purple>" + path + "</color>]";
            fromMetaJson = false;
            if (path.EndsWith("meta.json"))
            {
                fileName = "[<color=yellow>FROM META.JSON</color>] " + fileName;
                fromMetaJson = true;
            }
            jsNow = this.LoadJSON(path).AsObject;
            if (jsNow != null)
            {
                DrawUI();
            }
            else SuperController.LogError("PresetLoad: Can not load JSON.");
        }
        
        private void CheckCurrentScene()
        {
            lastPath = "CurrentScene";
            filePackageName = "";
            fileName = "[<color=purple>CurrentScene</color>]";
            fromMetaJson = false;
            jsNow = SuperController.singleton.GetSaveJSON();
            if (jsNow != null)
            {
                DrawUI();
            }
        }
        private void RescanCurrent()
        {
            if (lastPath == "") return;
            if (lastPath == "CurrentScene") CheckCurrentScene();
            else LoadJson(lastPath);
        }
        private void TracePath(JSONNode js2, string prefKey = "")
        {
            if (js2 == null) return;

            JSONClass js = js2.AsObject;
            if (js != null)
            {
                string id = js["id"].Value;
                if (js["id"] != null && id != string.Empty)
                {
                    if (prefKey.EndsWith("atoms[] "))
                        prefKey += "[id : <color=blue>" + id + "</color>] \n";
                    else
                    {
                        if (id.StartsWith("Plugin") || id.StartsWith("plugin#"))
                            prefKey += "[id : <color=purple>" + id + "</color>] \n";
                        else if(id == "asset")
                            prefKey += "[id : <color=#840956>" + id + "</color>] \n";
                        else if (id.Contains("Material") || id == "textures" || id == "sclera" || id == "irises" | id == "lacrimals" || id == "teeth" || id == "tongue" || id == "mouth" || id == "FemaleEyelashes" || id == "MaleEyelashes")
                            prefKey += "[id : <color=#a1488c>" + id + "</color>] \n";
                        else if (id.Contains(":/"))
                        {
                            string[] ids = id.Split(new string[] { ":/" }, StringSplitOptions.None);
                            prefKey += "[id : <color=red>" + ids[0] + "</color>] \n";
                        }
                        else
                            prefKey += "[id : " + id + "] \n";
                    }
                }
                if (js["AnimationName"] != null && js["AnimationName"].Value != string.Empty) prefKey += "[AnimationName : " + js["AnimationName"].Value + "] \n";
                
                foreach (string key in js.Keys)
                {
                    if (js[key].Value != string.Empty)
                    {
                        if (js[key].Value.Contains(":/"))
                        {
                            string suf = "";
                            if (prefKey.Contains("morphs[]"))
                            {
                                prefKey = prefKey.Replace("morphs[]", "<color=#930c5f>morphs</color>[]");
                                suf = " \n[name : <color=#930c5f>" + js["name"].Value + "</color>]";
                            }
                            if (prefKey.Contains("clothing[]")) 
                            {
                                prefKey = prefKey.Replace("clothing[]", "<color=#930c5f>clothing</color>[]");
                                suf = "[internalId : <color=#930c5f>" + js["internalId"].Value + "</color>]";
                            }
                            if (prefKey.Contains("hair[]"))
                            {
                                prefKey = prefKey.Replace("hair[]", "<color=#930c5f>clothing</color>[]");
                                suf = "[internalId : <color=#930c5f>" + js["internalId"].Value + "</color>]";
                            }

                            string name = js[key].Value.Split(new string[] { ":/" }, StringSplitOptions.None)[0];

                            bool flag = false;
                            string g = name;
                            if (g == "SELF" || g == selfName) continue;
                            if (g.Contains(':'))
                            {
                                string[] s0 = g.Split(':');
                                if (s0.Length == 2)
                                {
                                    g = s0[1];
                                }
                                else continue;
                            };
                            string[] s1 = g.Split('.');
                            if (s1.Length != 3) continue;
                            string packageGroupUid = s1[0] + '.' + s1[1];

                            string[] sTemp;
                            foreach (string n in packages)
                            {
                                if (GetPacGroupUid(n).ToLower() == packageGroupUid.ToLower())
                                {
                                    sTemp = n.Split('.');
                                    packageGroupUid = GetPacGroupUid(n);
                                    g = packageGroupUid + '.' + s1[2];
                                    s1[0] = sTemp[0];
                                    s1[1] = sTemp[1];
                                    break;
                                }
                            }
                            foreach (string nm in paths.Keys.ToArray())
                            {
                                if (nm.Contains(packageGroupUid))
                                {
                                    g = nm;
                                    flag = true;
                                    break;
                                }
                            }
                            if (!flag)
                            {
                                List<string> names2 = new List<string>();
                                foreach (string n in packages)
                                {
                                    if (GetPacGroupUid(n) == packageGroupUid) names2.Add(n);
                                }
                                if (names2.Count > 0 )
                                {
                                    if (!FileManagerSecure.FileExists("AddonPackages\\" + g + ".var"))
                                    {
                                        int version = 0;
                                        foreach (string n in names2)
                                        {
                                            try
                                            {
                                                int v = Int32.Parse(n.Split('.')[2]);
                                                if (version < v) version = v;
                                            }
                                            catch (Exception)
                                            {
                                                SuperController.LogMessage("TracePath: package name has wrong version format: " + n);
                                            }
                                        }
                                        if (version != 0)
                                        {
                                            if (g.EndsWith(".latest"))
                                            {
                                                g = "[<color=green>latest</color>] " + packageGroupUid + '.' + version;
                                            }
                                            else if (s1[2] != version.ToString())
                                            {
                                                //SuperController.LogMessage("AddonPackages\\" + g + ".var");
                                                g = "[<color=purple>V." + s1[2] + "->" + version + "</color>] " + packageGroupUid + '.' + version;
                                            }
                                        }
                                        else
                                        {
                                            g = "[<color=purple>NOTFOUND</color>] " + g;
                                        }
                                    }
                                }
                                else
                                {
                                    g = "[<color=purple>NOTFOUND</color>] " + g;
                                }
                            }

                            if (!paths.ContainsKey(g)) paths.Add(g, new List<string>());
                            if (key == "id")
                            {
                                paths[g].Add(prefKey + suf);
                            }
                            else
                            {
                                paths[g].Add(prefKey + "[" + key + " : <color=red>" + name + "</color>]" + suf);
                            }
                        }
                    }
                    else
                    {
                        TracePath(js[key], prefKey + key + "[] ");
                    }
                }
            }
            else
            {
                for (int i = 0; i < js2.Count; i++)
                {
                    TracePath(js2[i], prefKey);
                }
            }
        }
        private void DrawUI()
        {
            try
            {
                packages = FileManagerSecure.GetFiles("AddonPackages");
                scanned = false;
                wait = 0;
                selfName = "";
                if (SuperController.singleton != null && SuperController.singleton.LoadedSceneName != null)
                    selfName = SuperController.singleton.LoadedSceneName.Split(':')[0];
                paths.Clear();
                packageDownloader.ClearTrackedPackages();
                foreach (var item in varNames)
                {
                    if (item.Value.uiNew != null)
                        Destroy(item.Value.uiNew.gameObject);
                }
                if (myHubDownloader != null)
                {
                    foreach (var item in myHubDownloader.packagesContainer.GetComponents<HubResourcePackageUI>())
                    {
                        Destroy(item);
                    }
                }
                varNames.Clear();

                List<string> list = new List<string>();
                if (fromMetaJson)
                {
                    allDep = "";
                    AddDep(jsNow, "");
                }
                else
                {
                    TracePath(jsNow);
                    foreach (string keyN in paths.Keys.ToArray())
                    {
                        list.Add(keyN);
                    }
                    list.Sort();

                    allDep = "";
                    string key;
                    string license;
                    bool isLatest;
                    bool havePackage = true;
                    for (int i = 0; i < list.Count; i++)
                    {
                        havePackage = true;
                        isLatest = false;
                        key = list[i];
                        if (key.Contains("[<color=green>latest</color>] "))
                        {
                            isLatest = true;
                            key = key.Replace("[<color=green>latest</color>] ", string.Empty);
                            paths.Add(key, paths[list[i]]);
                            paths.Remove(list[i]);
                            list[i] = key;
                        }

                        JSONClass js2 = null;
                        string add = "";
                        if (!key.Contains("NOTFOUND"))
                        {
                            try
                            {
                                if (key.StartsWith("[<color"))
                                {
                                    havePackage = false;
                                    add = "V";
                                    js2 = this.LoadJSON(key.Split(new string[] { "] " }, StringSplitOptions.None)[1] + ":/meta.json").AsObject;
                                }
                                else
                                    js2 = this.LoadJSON(key + ":/meta.json").AsObject;
                            }
                            catch (Exception)
                            {
                                js2 = null;
                                add = "[<color=purple>ERROR</color>]";
                            }
                        }
                        else
                            havePackage = false;

                        if (js2 != null)
                        {
                            license = Colorize(js2["licenseType"]);
                            allDep += license + key + "\n";
                            //if (add != "V") AddDep(js2, "   ");
                        }
                        else
                        {
                            license = "[<color=red>MISSING</color>]";
                            allDep += license + add + key + "\n";
                        }
                        if (!varNames.ContainsKey(key))
                            varNames.Add(key, new VarItem { license = license, isLatest = isLatest, havePackage = havePackage });
                        else
                        {
                            varNames.Remove(key);
                            Dictionary<string, VarItem> vN = new Dictionary<string, VarItem>();
                            foreach (var item in varNames)
                            {
                                vN.Add(item.Key, item.Value);
                            }
                            varNames.Clear();
                            foreach (var item in vN)
                            {
                                varNames.Add(item.Key, item.Value);
                            }
                            varNames.Add(key, new VarItem { license = license, isLatest = isLatest, havePackage = havePackage });
                        }

                        if (js2 != null && add != "V")
                            AddDep(js2, "   ");

                    }
                    list.Sort();
                }
                
                List<string> list2 = new List<string> { "All Dependencies" };
                list2.AddRange(list);
                deps.choices = null;
                deps.choices = list2;
                deps.val = "All Dependencies";

                string clearName;
                string[] spl;
                string[] spl2;
                List<VarItem> list3 = new List<VarItem>();
                VarItem vi;
                foreach (var item in varNames)
                {
                    spl = item.Key.Split(new string[] { "] " }, StringSplitOptions.None);
                    if (spl.Length > 1)
                    {
                        if (spl[0].Contains("V."))
                        {
                            try
                            {
                                spl2 = spl[0].Split('.');
                                spl2 = spl2[1].Split('-');
                                int v = Int32.Parse(spl2[0]);
                                spl2 = spl[1].Split('.');
                                clearName = spl2[0] + "." + spl2[1] + "." + v;
                            }
                            catch (Exception e)
                            {
                                SuperController.LogMessage("DrawUI: Error while replacing version number " + spl[0] + " Name: " + spl[1] + "\n" + e );
                                clearName = spl[1];
                            }
                        }
                        else
                            clearName = spl[1];
                    }
                    else
                        clearName = spl[0];
                    if (item.Value.isLatest)
                    {
                        spl = clearName.Split('.');
                        if (spl.Length != 3) continue;
                        clearName = spl[0] + '.' + spl[1] + ".latest";
                    }
                    vi = item.Value;
                    vi.name = clearName;
                    list3.Add(vi);
                }
                varNames.Clear();
                foreach (VarItem item in list3)
                {
                    //SuperController.LogMessage("NAME " + item.name+" "+item.license+" "+item.isLatest);
                    varNames.Add(item.name, item);
                }

                allDep = "[<color=purple>TOTAL(include repeats/unique)</color> : " + (allDep.Split('\n').Length - 1) +"/"+ list3.Count + "] " + fileName + "\n" + allDep;
                tfUI.text = allDep;
            }
            catch (Exception e)
            {
                SuperController.LogMessage("DrawUI " + e);
            }
        }

        private void SelectDependency(string v)
        {
            try
            {
                if (v == "All Dependencies")
                {
                    tfUI.text = allDep;
                }
                else
                {
                    tfUI.text =  "";
                    Stack stc = new Stack();
                    for (int i = 0; i < paths[v].Count; i++)
                    {
                        string rep = paths[v][i].Replace("\n", string.Empty);
                        string[] st = rep.Split(new string[] { "] " }, StringSplitOptions.None);
                        for (int j = 0; j < st.Length-1; j++)
                        {
                            st[j] += "]";
                        }
                        stc = JSGet2(stc, st, 0);
                    }
                    string t1 = StrGet2(stc, "1");
                    int t1w = t1.Length;
                    if (t1.Length > 11000)
                    {
                        t1 = t1.Substring(0, 11000);
                        t1 = t1.Substring(0,t1.LastIndexOf('\n')) + "\nOVERFLOW: shown " + t1.LastIndexOf('\n') + " symbols from " + t1w;
                    }
                    tfUI.text += t1;
                }
            }
            catch (Exception e)
            {
                SuperController.LogMessage("SelectDependency " + e);
            }
        }
        private class Stack
        {
            public Dictionary<string,Stack> s;
            public Stack()
            {
                s = new Dictionary<string, Stack>();
            }
        }

        private Stack JSGet2(Stack stc, string[] st, int i)
        {
            if (st.Length > i)
            {
                if (!stc.s.ContainsKey(st[i]))
                {
                    stc.s[st[i]] = new Stack();
                }
                stc.s[st[i]] = JSGet2(stc.s[st[i]], st, i + 1);
            }
            return stc;
        }

        private string StrGet2(Stack stc, string pref)
        {
            string s = "";
            if (stc.s.Count >= 1)
            {
                pref += " ";
                s += "\n";
            }
            foreach (var v in stc.s)
            {
                if (v.Key.EndsWith("[]") && !s.EndsWith("\n")) s += "\n";
                if (pref == "1 ") { pref = ""; s = ""; }
                string s1 = v.Key;
                s += pref + s1;
                if (v.Value != null)
                {
                    s += pref + StrGet2(v.Value, pref);
                }
            }
            if (stc.s.Count == 0) s += "\n";
            return s;
        }

        private void AddDep(JSONClass js, string pref)
        {
            if (js != null)
            {
                JSONClass asObject = js["dependencies"].AsObject;
                if (asObject != null)
                {
                    bool isLatest;
                    bool havePackage = true;
                    foreach (string key in asObject.Keys)
                    {
                        havePackage = true;
                        isLatest = false;
                        string g = key;
                        string[] s1 = g.Split('.');
                        if (s1.Length != 3) continue;
                        string packageGroupUid = s1[0] + '.' + s1[1];

                        string[] sTemp;
                        foreach (string n in packages)
                        {
                            if (GetPacGroupUid(n).ToLower() == packageGroupUid.ToLower())
                            {
                                sTemp = n.Split('.');
                                packageGroupUid = GetPacGroupUid(n);
                                g = packageGroupUid + '.' + s1[2];
                                s1[0] = sTemp[0];
                                s1[1] = sTemp[1];
                                break;
                            }
                        }
                        List<string> names2 = new List<string>();
                        foreach (string n in packages)
                        {
                            if (GetPacGroupUid(n) == packageGroupUid) names2.Add(n);
                        }
                        if (names2.Count > 0)
                        {
                            if (!FileManagerSecure.FileExists("AddonPackages\\" + g + ".var"))
                            {
                                int version = 0;
                                foreach (string n in names2)
                                {
                                    try
                                    {
                                        int v = Int32.Parse(n.Split('.')[2]);
                                        if (version < v) version = v;
                                    }
                                    catch (Exception)
                                    {
                                        SuperController.LogMessage("AddDep: package name has wrong version format: " + n);
                                    }
                                }
                                if (version != 0)
                                {
                                    if (g.EndsWith(".latest"))
                                    {
                                        g = packageGroupUid + '.' + version;
                                        isLatest = true;
                                    }
                                    else if (s1[2] != version.ToString())
                                    {
                                        havePackage = false;
                                        //SuperController.LogMessage("AddonPackages\\" + g + ".var");
                                        g = "[<color=purple>V." + s1[2] + "->" + version + "</color>] " + packageGroupUid + '.' + version;
                                    }
                                }
                                else
                                {
                                    havePackage = false;
                                    g = "[<color=purple>NOTFOUND</color>] " + g;
                                }
                                

                            }
                        }
                        else
                        {
                            havePackage = false;
                            g = "[<color=purple>NOTFOUND</color>] " + g;
                        }

                        string license;
                        JSONClass js2 = null;
                        string add = "";
                        if (fromMetaJson)
                        {
                            license = Colorize(asObject[key]["licenseType"]);
                            allDep += pref + license + g + "\n";
                            //AddDep(asObject[key].AsObject, pref + "   ");
                        }
                        else
                        {
                            if (!g.Contains("NOTFOUND"))
                            {
                                try
                                {
                                    if (g.StartsWith("[<color"))
                                    {
                                        add = "V";
                                        js2 = this.LoadJSON(g.Split(new string[] { "] " }, StringSplitOptions.None)[1] + ":/meta.json").AsObject;
                                    }
                                    else
                                        js2 = this.LoadJSON(g + ":/meta.json").AsObject;
                                }
                                catch (Exception)
                                {
                                    js2 = null;
                                    add = "[<color=purple>ERROR</color>]";
                                }
                            }
                            if (js2 != null)
                            {
                                license = Colorize(js2["licenseType"]);
                                allDep += pref + license + g + "\n";
                                //if (add != "V") AddDep(js2, pref + "   ");
                            }
                            else
                            {
                                license = "[<color=red>MISSING</color>]";
                                allDep += pref + license + add + g + "\n";
                            }
                        }
                        if (!varNames.ContainsKey(g))
                            varNames.Add(g, new VarItem { license = license, isLatest = isLatest, havePackage = havePackage, isSubDependency = (pref != "") });
                        else if (pref == "")
                        {
                            varNames.Remove(g);
                            Dictionary<string, VarItem> vN = new Dictionary<string, VarItem>();
                            foreach (var item in varNames)
                            {
                                vN.Add(item.Key, item.Value);
                            }
                            varNames.Clear();
                            foreach (var item in vN)
                            {
                                varNames.Add(item.Key, item.Value);
                            }
                            varNames.Add(g, new VarItem { license = license, isLatest = isLatest, havePackage = havePackage });
                        }

                        if (fromMetaJson)
                            AddDep(asObject[key].AsObject, pref + "   ");
                        else if(js2 != null && add != "V")
                            AddDep(js2, pref + "   ");
                    }
                }
            }
        }

        private string Colorize(string s)
        {
            if (s.Contains("PC") || s.Contains("NC") || s.Contains("estionab") || s.Contains("MISSING") || s == "null")
            {
                s = "[<color=red>" + s + "</color>] ";
            }
            else if (s.Contains("BY-ND"))
            {
                s = "[<color=yellow>" + s + "</color>] ";
            }
            else if (s == "CC BY-SA" || s == "CC BY" || s == "FC")
            {
                s = "[<color=green>" + s + "</color>] ";
            }
            else
            {
                s = "[" + s + "] ";
            }
            return s;
        }

        protected Transform CreateFullWidthUIElement(Transform prefab)
        {
            Transform transform = null;
            if (prefab != null)
            {
                transform = Instantiate(prefab);
                if (fullWidthUIContent != null)
                {
                    //float offsetY = leftUIContent.sizeDelta.y;
                    //if (rightUIContent.sizeDelta.y > offsetY) offsetY = rightUIContent.sizeDelta.y;
                    //SuperController.LogMessage("left "+ leftUIContent.sizeDelta.y+" right "+ rightUIContent.sizeDelta.y);
                    fullWidthUIContent.anchoredPosition = new Vector2(fullWidthUIContent.anchoredPosition.x, -125);
                    fullWidthUIContent.transform.SetAsFirstSibling();
                    transform.SetParent(fullWidthUIContent, worldPositionStays: false);
                }
                transform.gameObject.SetActive(value: true);
                fullWidthUIElements.Add(transform);
            }
            return transform;
        }

        public UIDynamicTextField CreateFullWidthTextField(JSONStorableString jss)
        {
            UIDynamicTextField uIDynamicTextField = null;
            if (manager != null && manager.configurableTextFieldPrefab != null && jss.text == null)
            {
                Transform transform = CreateFullWidthUIElement(manager.configurableTextFieldPrefab.transform);
                if (transform != null)
                {
                    uIDynamicTextField = transform.GetComponent<UIDynamicTextField>();
                    if (uIDynamicTextField != null)
                    {
                        jss.dynamicText = uIDynamicTextField;
                    }
                }
            }
            return uIDynamicTextField;
        }
        public UIDynamicPopup CreateFullWidthFilterablePopup(JSONStorableStringChooser jsc)
        {
            UIDynamicPopup uIDynamicPopup = null;
            if (manager != null && manager.configurableFilterablePopupPrefab != null && jsc.popup == null)
            {
                Transform transform = CreateFullWidthUIElement(manager.configurableFilterablePopupPrefab.transform);
                if (transform != null)
                {
                    uIDynamicPopup = transform.GetComponent<UIDynamicPopup>();
                    if (uIDynamicPopup != null)
                    {
                        uIDynamicPopup.label = jsc.name;
                        jsc.popup = uIDynamicPopup.popup;
                    }
                }
            }

            return uIDynamicPopup;
        }
        public UIDynamicButton CreateFullWidthButton(string label)
        {
            UIDynamicButton uIDynamicButton = null;
            if (manager != null && manager.configurableButtonPrefab != null)
            {
                Transform transform = CreateFullWidthUIElement(manager.configurableButtonPrefab.transform);
                if (transform != null)
                {
                    uIDynamicButton = transform.GetComponent<UIDynamicButton>();
                    if (uIDynamicButton != null)
                    {
                        uIDynamicButton.label = label;
                    }
                }
            }

            return uIDynamicButton;
        }
        void OnDestroy()
        {
            packageDownloader.ClearTrackedPackages();
            if(myPackageDownloaderUIObject != null) Destroy(myPackageDownloaderUIObject);
        }

	}

}